/*
  Copyright (C) 1999  Andrew J. Bettison

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

   For author contact information, look in the README.bscanf file.
*/

#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>
#include <string.h>
#include <ctype.h>

/*
	bscanf(), vbscanf()	-- Version 1.01, 31-Oct-1999

	An interface to sscanf(3) that advances a pointer through a buffer,
	instead of just consuming chars from a buffer.

	We process the 'fmt' string one conversion at a time, and delegate
	each conversion to sscanf(3).  This means parsing the 'fmt' string
	to find all the "%..." sequences.  We append a "%n" to each one to
	get sscanf(3) to tell us how many characters it consumes, and that's
	how far to advance the pointer after each call to sscanf(3).  We
	deal with characters which are not part of any "%..." sequence
	ourselves because it's easy.

	But wait!  Couldn't we just use one call to sscanf(3) and append
	"%n" to the whole 'fmt' string (being careful of any unmatched '%'
	at the end)?
	
	No.  The problem is that there is no way for us to append our own
	'&nchars' argument to the end of a va_alist, because we don't know
	where it ends!  We would have to analyse the 'fmt' string in detail,
	which is what we are doing anyway.
	
	So why bother?  Just use "%n" every time you call sscanf(3).

	Sure.  But this function is designed to fill the need where there is
	already existing code that calls sscanf() lots of times, and we
	don't want to edit every single call to sscanf(), except to change
	it to bscanf() and change the type of the first argument... much
	easier.
	
	-- Andrew Bettison <andrewb@zip.com.au>, 31 October, 1999
*/

int
vbscanf(const char **buf, const char *fmt, va_list ap)
{
	const char *bufs = *buf;	/* remember the start of the buffer */
	int ret = 0;
	const char *f;
	const char *sfmt;
	enum { small, normal, big, very_big, bad } size = bad;
	int charset = 0;
	int suppress = 0;
	int alloc = 0;

	/* As defined in the man page of sscanf(3).  */
	if (!*(*buf))
		return EOF;

	for (sfmt = NULL, f = fmt; *f; f++)
	{
		if (!sfmt)
		{
			/* Handle chars that are not part of a "%..." sequence.  If
			   we find a '%', then we start a sequence by setting the
			   'sfmt' pointer.  Otherwise, we just literally match each
			   character, treating white space specially.
			 */
			if (*f == '%')
			{
				sfmt = f;
				size = normal;
				charset = 0;
				suppress = 0;
				alloc = 0;
				size = normal;
			}
			else if (isspace(*f))
			{
				while (*(*buf) && isspace(*(*buf)))
					(*buf)++;
			}
			else if (*(*buf) == *f)
				(*buf)++;
			else
				break; /* for */
		}
		else
		{
			if (!charset && *f == '%')
			{
				/* We've found a "%%" sequence (or "%...%").  That just
				   requires the next input char to be a single '%'.
				 */
				if (*(*buf) != '%')
					break; /* for */
				else
					(*buf)++;
			}
			else if (charset ? *f == ']' : !!strchr("dDiouxXfegEscpn", *f))
			{

				/* We've found a complete "%..." conversion sequence.  */

				if (size != bad)
				{
					int nchar = 0;
					char conv = *f;
					char *tfmt;

					/* Make a copy of the "%..." fmt string with "%n" appended.  */
					tfmt = malloc(f - sfmt + 5);
					if (tfmt == NULL)
						abort();
					strncpy(tfmt, sfmt, f - sfmt + 1);
					tfmt[f - sfmt + 1] = '\0';
					if (*f == 'D' && size == normal)
					{
						tfmt[f - sfmt] = 'l';
						tfmt[f - sfmt + 1] = conv = 'd';
						tfmt[f - sfmt + 2] = '\0';
						size = big;
					}
					strcat(tfmt, "%n");

					switch (conv)
					{
					case 'd':
					case 'i':
					case 'o':
					case 'u':
					case 'x':
					case 'X':
						/* Can't use 'a' flag for numeric conversions.  */
						if (alloc)
							break;
						if (suppress)
							ret += sscanf((*buf), tfmt, &nchar);
						else
							switch (size)
							{
							case small:
								ret += sscanf((*buf), tfmt, va_arg(ap, short *), &nchar);
								break;
							case big:
								ret += sscanf((*buf), tfmt, va_arg(ap, long *), &nchar);
								break;
							case very_big:
								ret += sscanf((*buf), tfmt, va_arg(ap, long long *), &nchar);
								break;
							default:
								ret += sscanf((*buf), tfmt, va_arg(ap, int *), &nchar);
								break;
							}
						sfmt = NULL;
						break;

					case 'f':
					case 'e':
					case 'g':
					case 'E':
						/* Can't use 'a' or 'h' flags for floating-point
						   conversions.
						 */
						if (alloc || size == small)
							break;
						if (suppress)
							ret += sscanf((*buf), tfmt, &nchar);
						else
							switch (size)
							{
							case big:
								ret += sscanf((*buf), tfmt, va_arg(ap, double *), &nchar);
								break;
							case very_big:
								ret += sscanf((*buf), tfmt, va_arg(ap, long double *), &nchar);
								break;
							default:
								ret += sscanf((*buf), tfmt, va_arg(ap, float *), &nchar);
								break;
							}
						sfmt = NULL;
						break;

					case 's':
					case 'c':
					case ']':
						/* Can't use 'h', 'l', 'L', or 'q' flags for string
						   conversions.
						 */
						if (size != normal)
							break;
						if (suppress)
							ret += sscanf((*buf), tfmt, &nchar);
						else if (alloc)
							ret += sscanf((*buf), tfmt, va_arg(ap, char **), &nchar);
						else
							ret += sscanf((*buf), tfmt, va_arg(ap, char *), &nchar);
						sfmt = NULL;
						break;

					case 'p':
						/* Can't use 'a', 'h', 'l', 'L', or 'q' flags for pointer
						   conversion.
						 */
						if (alloc || size != normal)
							break;
						if (suppress)
							ret += sscanf((*buf), tfmt, &nchar);
						else
							ret += sscanf((*buf), tfmt, va_arg(ap, void *), &nchar);
						sfmt = NULL;
						break;

					case 'n':
						/* Can't use 'a' flag for "%n" conversion.  */
						if (alloc)
							break;
						if (suppress)
							ret += sscanf((*buf), tfmt, &nchar);
						else
							switch (size)
							{
							case small:
								*va_arg(ap, short *) = (*buf) - bufs;
								break;
							case big:
								*va_arg(ap, long *) = (*buf) - bufs;
								break;
							case very_big:
								*va_arg(ap, long long *) = (*buf) - bufs;
								break;
							default:
								*va_arg(ap, int *) = (*buf) - bufs;
								break;
							}
						sfmt = NULL;
						break;
					}

					/* This is the whole point of the exercise!  */
					(*buf) += nchar;

					free(tfmt);
				}

				/* If 'sfmt' is still non-NULL, it means that the code
				   above was not happy with the syntax of the "%..."
				   sequence, so it never called sscanf().  Instead, we
				   will treat the sequence as literal characters that
				   must match explicitly.  Possibly the real sscanf() is
				   more permissive than we are about irrelevant flag
				   chars like 'h' or 'l' being used with conversion
				   characters like 'e' or 's', but it's better to be
				   conservative because otherwise we could end up
				   passing a "%..." sequence to sscanf() that it doesn't
				   accept, and popping an argument using va_arg() where
				   sscanf wouldn't, and then everything would be a mess!
				 */
				if (sfmt)
				{
					for (; sfmt != f; sfmt++)
						if (isspace(*sfmt))
							while (*(*buf) && isspace(*(*buf)))
								(*buf)++;
						else if (*(*buf) == *sfmt)
							(*buf)++;
						else
							break; /* for */
					sfmt = NULL;
				}
			}
			else if (!charset)
			{
				if (*f == '[')
				{
					/* We've found a "%[...]" conversion.  We set 'charset'
					   to consume all chars up to the matching ']',
					   taking care to allow the sequences "%[]...]" and
					   "%[^]...]".
					 */
					charset = 1;
					if (f[1] == ']')
						f++;
					else if (f[1] == '^' && f[2] == ']')
						f += 2;
				}
				else if (strchr("*ahlLq0123456789", *f))
				{
					/* Consume legal flag characters.  We indicate
					   contradictory flags by setting size = bad.
					 */
					switch (*f)
					{
					case '*':
						suppress = 1;
						break;
					case 'a':
						alloc = 1;
						break;
					case 'h':
						if (size == normal)
							size = small;
						else
							size = bad;
						break;
					case 'l':
						if (size == normal)
							size = big;
						else if (size == big)
							size = very_big;
						else
							size = bad;
						break;
					case 'L':
					case 'q':
						if (size == normal)
							size = very_big;
						else
							size = bad;
						break;
					}
				}
			}
		}
	}

	return ret;
}

int
bscanf(const char **buf, const char *fmt, ...)
{
	va_list ap;
	int ret;

	va_start(ap, fmt);
	ret = vbscanf(buf, fmt, ap);
	va_end(ap);
	return ret;
}
